# As setup.py imports this module to get the version, try not to do anything
#   with dependencies for the project here. If that happens, setup.py
#   should not import tornado_json and instead use this find_version
#   thing: https://github.com/jezdez/envdir/blob/a062497e4339d5eb11e8a95dc6186dea6231aeb1/setup.py#L24
# Alternatively, just put the version in a text file or something to avoid
#   this.

__version__ = '1.3.4'
